import java.awt.*;
import java.awt.event.*;
import acm.graphics.*;

public class ToneMatrixDisplay extends GCanvas implements MouseListener, MouseMotionListener, ComponentListener {	
	/* Number of pixels on the screen we'd like each pixel to be. */
	private static final int GSCALE = 20;
	
	/* Colors for lit and unlit. */
	private static final Color LIT_COLOR = new Color(232, 232, 232);
	private static final Color UNLIT_COLOR = new Color(134, 110, 255).darker();
	
	/* The color of the highlight rectangle. */
	private static final Color HIGHLIGHT_RECT_COLOR = new Color(255, 255, 0, 128);
	
	/* The background color (not on the matrix itself. */
	private static final Color BACKGROUND_COLOR = Color.GRAY;
	
	/* The values in the tone matrix. */
	private boolean[][] matrix = new boolean[ToneMatrixConstants.size()][ToneMatrixConstants.size()];
	
	/* What's displayed: the rectangles representing those matrix elements. */
	private GRect[][] matrixRects = new GRect[ToneMatrixConstants.size()][ToneMatrixConstants.size()];
	
	/* Scaling factors determining the size of each box. */
	private double xScale, yScale;
	
	/* X and Y translation to the start of the matrix. */
	private double xTrans, yTrans;	
	
	/* Whether we're setting or clearing during a drag operation. */
	private boolean isMarking;
	
	/* A rectangle that highlights which column is playing. */
	private GRect highlightRect;
	
	/* An indicator of which column we're in.  A value of -1 indicates that we are
	 * not anywhere.  This is marked volatile because the MIDI thread might change
	 * this value with no intervening reads.
	 */
	private volatile int highlightedColumn = -1;
		
	public ToneMatrixDisplay() {
		setBackground(BACKGROUND_COLOR);
		
		/* Initialize all of the blocks in the tone matrix. */
		for (int i = 0; i < ToneMatrixConstants.size(); i++) {
			for (int j = 0; j < ToneMatrixConstants.size(); j++) {
				matrixRects[i][j] = new GRect(0, 0, 0, 0);
				matrixRects[i][j].setFilled(true);
				matrixRects[i][j].setColor(Color.BLACK);
				matrixRects[i][j].setFillColor(UNLIT_COLOR);
				
				add(matrixRects[i][j]);
			}
		}
		
		/* Create the highlighting rectangle. */
		highlightRect = new GRect(0, 0, 0, 0);
		highlightRect.setFilled(true);
		highlightRect.setColor(HIGHLIGHT_RECT_COLOR);
		add(highlightRect);
		
		/* Try to make us have a nice size. */
		setPreferredSize(new Dimension(ToneMatrixConstants.size() * GSCALE, ToneMatrixConstants.size() * GSCALE));
		
		/* Register listeners for resizing and mouse events. */
		addMouseListener(this);
		addMouseMotionListener(this);
		addComponentListener(this);
	}
	
	/**
	 * Redraws the display.
	 */
	private void redrawAll() {
		/* Update all squares. */
		for (int row = 0; row < ToneMatrixConstants.size(); row++) {
			for (int col = 0; col < ToneMatrixConstants.size(); col++) {
				matrixRects[row][col].setBounds(xScale * col + xTrans, yScale * row + yTrans, xScale, yScale);
				matrixRects[row][col].setFillColor(matrix[row][col]? LIT_COLOR : UNLIT_COLOR);
			}
		}
		
		/* Reposition the highlight rect. */
		if (highlightedColumn == -1) {
			highlightRect.setBounds(0, 0, 0, 0);
		} else {
			highlightRect.setBounds(xScale * highlightedColumn + xTrans, yTrans, xScale, yScale * ToneMatrixConstants.size());
		}
	}
	
	/**
	 * Recalculates all of the display geometry after a resize.
	 */
	private void recalculateGeometry() {
		/* Compute what the x and y scales would be. */
		xScale = getWidth()  / (double)ToneMatrixConstants.size();
		yScale = getHeight() / (double)ToneMatrixConstants.size();
		
		/* Pick the smaller of the two. */
		xScale = yScale = Math.min(xScale, yScale);
		
		/* Compute buffering. */
		xTrans = (getWidth() - xScale * ToneMatrixConstants.size()) / 2.0;
		yTrans = (getHeight() - yScale * ToneMatrixConstants.size()) / 2.0;
	}
	
	/**
	 * Returns whether (row, col) is in bounds.
	 * 
	 * @param row The row to check.
	 * @param col The column to check.
	 * @return Whether it is in bounds.
	 */
	private boolean inBounds(int row, int col) {
		return row >= 0 && col >= 0 && row < ToneMatrixConstants.size() && col < ToneMatrixConstants.size();
	}
	
	/**
	 * Draws a mark at the given location.
	 * 
	 * @param row The row to mark.
	 * @param col The column to mark.
	 * @param set Whether we're setting or clearing the mark
	 */
	private void markAt(int row, int col, boolean set) {
		if (inBounds(row, col)) {
			matrix[row][col] = set;
			matrixRects[row][col].setFillColor(set? LIT_COLOR : UNLIT_COLOR);
		}
	}
	
	/**
	 * Clears the display.
	 */
	public void clear() {
		for (int i = 0; i < ToneMatrixConstants.size(); i++) {
			for (int j = 0; j < ToneMatrixConstants.size(); j++) {
				markAt(i, j, false);
			}
		}
	}

	/**
	 * Colors the appropriate pixels black.
	 * 
	 * @param e The MouseEvent that triggered this.
	 */
	public void mouseDragged(MouseEvent e) {
		/* Make a mark here. */
		int x = (int)((e.getX() - xTrans) / xScale);
		int y = (int)((e.getY() - yTrans) / yScale);
		markAt(y, x, isMarking);
	}
	
	public void mousePressed(MouseEvent e) {
		/* Make a mark here. */
		int x = (int)((e.getX() - xTrans) / xScale);
		int y = (int)((e.getY() - yTrans) / yScale);
		
		/* Determine whether we're marking or clearing. */
		if (inBounds(y, x)) {
			isMarking = !matrix[y][x];
		}
		
		markAt(y, x, isMarking);
	}
	
	/* Required but unused methods */
	public void mouseClicked(MouseEvent arg0) {}
	public void mouseEntered(MouseEvent arg0) {}
	public void mouseExited(MouseEvent arg0) {}
	public void mouseReleased(MouseEvent arg0) {}
	public void mouseMoved(MouseEvent arg0) {}
	public void componentHidden(ComponentEvent arg0) {}
	public void componentMoved(ComponentEvent arg0) {}
	public void componentShown(ComponentEvent arg0) {}
	
	/* If we resize, we need to redraw everything. */
	public void componentResized(ComponentEvent arg0) {
		recalculateGeometry();
		redrawAll();
	}
	
	/**
	 * Returns a copy of the underlying matrix.
	 * 
	 * @return A copy of the underlying matrix.
	 */
	public boolean[][] getMatrix() {
		/* We store the matrix upside-down in order to make the math with cursor
		 * events easier to process.  This means that we have to flip the matrix
		 * before returning it.
		 */
		boolean[][] result = new boolean[ToneMatrixConstants.size()][ToneMatrixConstants.size()];
		for (int i = 0; i < ToneMatrixConstants.size(); i++) {
			for (int j = 0; j < ToneMatrixConstants.size(); j++) {
				result[ToneMatrixConstants.size() - 1 - i][j] = matrix[i][j];
			}
		}
		return result;
	}
	
	/**
	 * Stores the given matrix as the underlying matrix for the display.  The
	 * matrix must have the correct dimensions.
	 * 
	 * @param newMatrix The matrix to set.
	 */
	public void setMatrix(boolean[][] newMatrix) {
		/* Confirm this is a valid matrix. */
		if (newMatrix == null) throw new NullPointerException("setMatrix cannot be called on a null matrix.");
		if (newMatrix.length != ToneMatrixConstants.size()) throw new IllegalArgumentException("Incorrect size for setMatrix");
		for (int i = 0; i < newMatrix.length; i++) {
			if (newMatrix[i] == null)
				throw new IllegalArgumentException("Inner array is null.");
			if (newMatrix[i].length != ToneMatrixConstants.size()) throw new IllegalArgumentException("Incorrect size for setMatrix");
		}
		
		/* Flip the matrix before storing it. */
		for (int row = 0; row < ToneMatrixConstants.size(); row++) {
			for (int col = 0; col < ToneMatrixConstants.size(); col++) {
				matrix[ToneMatrixConstants.size() - 1 - row][col] = newMatrix[row][col];
			}
		}
		
		redrawAll();
	}
	
	/**
	 * Sets which column is currently highlighted.  If a value is given that is out
	 * of range, no column will be lit.
	 * 
	 * @param column The column that should be lit.
	 */
	public void highlightColumn(int column) {
		/* Clamp the value. */
		if (column < 0 || column >= ToneMatrixConstants.size())
			column = -1;
		
		highlightedColumn = column;
		redrawAll();
	}
}
